#!/usr/bin/python3

import gi
from gi.repository import Gio, GLib

import datetime
import gettext
import html
import json
import os
import subprocess
import sys
import time
import re
import threading

gettext.install("mintupdate", "/usr/share/locale")

# These updates take priority over other updates.
# If a new version of these packages is available, nothing else is listed.
PRIORITY_UPDATES = ['mintupdate', 'mint-upgrade-info']

settings = Gio.Settings(schema_id="com.linuxmint.updates")

SUPPORTED_KERNEL_TYPES = ["-generic", "-lowlatency", "-aws", "-azure", "-gcp", "-kvm", "-oem", "-oracle"]
KERNEL_PKG_NAMES = ['linux-headers-VERSION', 'linux-headers-VERSION-KERNELTYPE', 'linux-image-VERSION-KERNELTYPE', \
    'linux-modules-VERSION-KERNELTYPE', 'linux-modules-extra-VERSION-KERNELTYPE']
KERNEL_PKG_NAMES.append('linux-image-extra-VERSION-KERNELTYPE') # Naming convention in 16.04, until 4.15 series

CONFIGURED_KERNEL_TYPE = settings.get_string("selected-kernel-type")
if CONFIGURED_KERNEL_TYPE not in SUPPORTED_KERNEL_TYPES:
    CONFIGURED_KERNEL_TYPE = "-generic"

CONFIG_PATH = os.path.expanduser("~/.linuxmint/mintupdate")

# Used as a decorator to run things in the background
def _async(func):
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()
        return thread
    return wrapper

# Used as a decorator to run things in the main loop, from another thread
def _idle(func):
    def wrapper(*args):
        GLib.idle_add(func, *args)
    return wrapper

def get_release_dates():
    """ Get distro release dates for support duration calculation """
    release_dates = {}
    distro_info = []
    if os.path.isfile("/usr/share/distro-info/ubuntu.csv"):
        distro_info += open("/usr/share/distro-info/ubuntu.csv", "r").readlines()
    if os.path.isfile("/usr/share/distro-info/debian.csv"):
        distro_info += open("/usr/share/distro-info/debian.csv", "r").readlines()
    if distro_info:
        for distro in distro_info[1:]:
            try:
                distro = distro.split(",")
                release_date = time.mktime(time.strptime(distro[4], '%Y-%m-%d'))
                release_date = datetime.datetime.fromtimestamp(release_date)
                support_end = time.mktime(time.strptime(distro[5].rstrip(), '%Y-%m-%d'))
                support_end = datetime.datetime.fromtimestamp(support_end)
                release_dates[distro[2]] = [release_date, support_end]
            except:
                pass
    return release_dates

class KernelVersion():

    def __init__(self, version):
        field_length = 3
        self.version = version
        self.version_id = []
        version_id = self.version.replace("-", ".").split(".")
        # Check if mainline rc kernel to ensure proper sorting vs mainline release kernels
        suffix = next((x for x in version_id if x.startswith("rc")), None)
        if not suffix:
            suffix = "z"
        # Copy numeric parts from version_id to self.version_id and fill up to field_length
        for element in version_id:
            if element.isnumeric():
                self.version_id.append("0" * (field_length - len(element)) + element)
        # Installed kernels always have len(self.version_id) >= 4 at this point,
        # create missing parts for not installed mainline kernels:
        while len(self.version_id) < 3:
            self.version_id.append("0" * field_length)
        if len(self.version_id) == 3:
            self.version_id.append("%s%s" % (''.join((x[:field_length - 2].lstrip('0') + x[field_length - 2:] for x in self.version_id)), suffix))
        elif len(self.version_id[3]) == 6:
            # installed release mainline kernel, add suffix for sorting
            self.version_id[3] += suffix
        self.series = tuple(self.version_id[:3])
        self.shortseries = tuple(self.version_id[:2])

class Update():

    def __init__(self, package=None, source_name=None):
        self.package_names = []
        if package is not None:
            self.package_names.append(package.name)
            self.source_packages = {"%s=%s" % (package.candidate.source_name, package.candidate.source_version)}
            self.main_package_name = package.name
            self.package_name = package.name
            self.new_version = package.candidate.version
            if package.installed is None:
                self.old_version = ""
            else:
                self.old_version = package.installed.version
            self.size = package.candidate.size
            self.real_source_name = package.candidate.source_name
            if source_name is not None:
                self.source_name = source_name
            else:
                self.source_name = self.real_source_name
            self.display_name = self.source_name
            self.short_description = package.candidate.raw_description
            self.description = package.candidate.description
            self.archive = ""

            self.type = "package"
            self.origin = ""
            for origin in package.candidate.origins:
                self.origin = origin.origin
                self.site = origin.site
                self.archive = origin.archive
                if origin.origin == "Ubuntu":
                    self.origin = "ubuntu"
                elif origin.origin == "Debian":
                    self.origin = "debian"
                elif origin.origin.startswith("LP-PPA"):
                    self.origin = origin.origin
                if origin.origin == "Ubuntu" and '-security' in origin.archive:
                    self.type = "security"
                    break
                if origin.origin == "Debian" and '-Security' in origin.label:
                    self.type = "security"
                    break
                if source_name in ["firefox", "thunderbird", "chromium"]:
                    self.type = "security"
                    break
                if origin.origin == "linuxmint":
                    if origin.component == "romeo":
                        self.type = "unstable"
                break
            if package.candidate.section == "kernel" or self.package_name.startswith("linux-headers") or self.real_source_name in ["linux", "linux-kernel", "linux-signed", "linux-meta"]:
                self.type = "kernel"

    def add_package(self, pkg):
        self.package_names.append(pkg.name)
        self.source_packages.add("%s=%s" % (pkg.candidate.source_name, pkg.candidate.source_version))
        self.size += pkg.candidate.size
        if self.main_package_name is None or pkg.name == self.source_name:
            self.overwrite_main_package(pkg)
            return

        if self.main_package_name != self.source_name:
            # Overwrite dev, dbg, common, arch packages
            for suffix in ["-dev", "-dbg", "-common", "-core", "-data", "-doc", ":i386", ":amd64"]:
                if (self.main_package_name.endswith(suffix) and not pkg.name.endswith(suffix)):
                    self.overwrite_main_package(pkg)
                    return
            # Overwrite lib packages
            for prefix in ["lib", "gir1.2"]:
                if (self.main_package_name.startswith(prefix) and not pkg.name.startswith(prefix)):
                    self.overwrite_main_package(pkg)
                    return
            for keyword in ["-locale-", "-l10n-", "-help-"]:
                if (keyword in self.main_package_name) and (keyword not in pkg.name):
                    self.overwrite_main_package(pkg)
                    return

    def overwrite_main_package(self, pkg):
        self.description = pkg.candidate.description
        self.short_description = pkg.candidate.raw_description
        self.main_package_name = pkg.name

class Alias():
    def __init__(self, name, short_description, description):

        name = name.strip()
        short_description = short_description.strip()
        description = description.strip()

        if (name.startswith('_("') and name.endswith('")')):
            name = _(name[3:-2])
        if (short_description.startswith('_("') and short_description.endswith('")')):
            short_description = _(short_description[3:-2])
        if (description.startswith('_("') and description.endswith('")')):
            description = _(description[3:-2])

        self.name = name
        self.short_description = short_description
        self.description = description

class UpdateTracker():

    # Loads past updates from JSON file
    def __init__(self, settings, logger):
        os.system("mkdir -p %s" % CONFIG_PATH)
        self.path = os.path.join(CONFIG_PATH, "updates.json")

        self.test_mode = os.getenv("MINTUPDATE_TEST") == "tracker-max-age"

        self.tracker_version = 1 # version of the data structure
        self.settings = settings
        self.tracked_updates = {}
        self.refreshed_update_names = [] # updates which are seen in checkAPT
        self.today = datetime.date.today().strftime("%Y.%m.%d")
        self.max_days = 0 # oldest update (in number of days seen)
        self.oldest_since_date = self.today # oldest update (according to since date)
        self.active = True # False if the tracking was already done today
        self.security_only = self.settings.get_boolean("tracker-security-only")
        self.logger = logger

        try:
            with open(self.path) as f:
                self.tracked_updates = json.load(f)
                if self.tracked_updates['version'] < self.tracker_version:
                    raise Exception()
                if self.tracked_updates['checked'] > self.today:
                    raise Exception()
                if self.tracked_updates['notified'] > self.today:
                    raise Exception()
                if self.tracked_updates['checked'] == self.today:
                    # We already tracked updates today
                    self.active = False
        except Exception as e:
            self.logger.write("Tracker exception: " + str(e))
            self.tracked_updates['updates'] = {}
            self.tracked_updates['version'] = self.tracker_version
            self.tracked_updates['checked'] = self.today
            self.tracked_updates['notified'] = self.today

    # Updates the record for a particular update
    def update(self, update):
        self.refreshed_update_names.append(update.real_source_name)
        if update.real_source_name not in self.tracked_updates['updates']:
            update_record = {}
            update_record['type'] = update.type
            update_record['since'] = self.today
            update_record['days'] = 1
            self.tracked_updates['updates'][update.real_source_name] = update_record
        else:
            update_record = self.tracked_updates['updates'][update.real_source_name]
            update_record['type'] = update.type
            if self.today > self.tracked_updates['checked']:
                update_record['days'] += 1

        if update.type in ["security", "kernel"] or (not self.security_only):
            if self.max_days < update_record['days']:
                self.max_days = update_record['days']
            if self.oldest_since_date > update_record['since']:
                self.oldest_since_date = update_record['since']

    # Returns the number of days between today and the given date string
    def get_days_since_date(self, string: str, date_format: str) -> int:
        if string is None:
            return 999
        datetime_object = datetime.datetime.strptime(string, date_format)
        days = (datetime.date.today() - datetime_object.date()).days
        return days

    # Returns the number of days between today and the given timestamp
    def get_days_since_timestamp(self, timestamp: float) -> int:
        if timestamp == 0:
            return 999
        datetime_object = datetime.datetime.fromtimestamp(timestamp)
        days = (datetime.date.today() - datetime_object.date()).days
        return days

    def get_latest_apt_upgrade(self):
        latest_upgrade_date = None

        if os.path.exists("/var/log/apt/history.log"):
            logs = subprocess.getoutput("cat /var/log/apt/history.log")
            for event in logs.split("\n\n"):
                if "Upgrade: " not in event:
                    continue
                end_date = None
                for line in event.split("\n"):
                    line = line.strip()
                    if line.startswith("End-Date: "):
                        end_date = line.replace("End-Date: ", "")
                        end_date = end_date.split()[0]
                if end_date is not None and (latest_upgrade_date is None or end_date > latest_upgrade_date):
                    latest_upgrade_date = end_date

        if latest_upgrade_date is None:
            try:
                logs = subprocess.getoutput("zcat /var/log/apt/history.log*gz")
                for event in logs.split("\n\n"):
                    if "Upgrade: " not in event:
                        continue
                    end_date = None
                    for line in event.split("\n"):
                        line = line.strip()
                        if line.startswith("End-Date: "):
                            end_date = line.replace("End-Date: ", "")
                            end_date = end_date.split()[0]
                    if end_date is not None and (latest_upgrade_date is None or end_date > latest_upgrade_date):
                        latest_upgrade_date = end_date
            except Exception as e:
                print("Failed to check compressed APT logs", e)

        return latest_upgrade_date

    # Returns true if a notification is required and updates the tracker
    # with the new notification date
    def notify(self):
        # Check notification enabled
        if self.settings.get_boolean("tracker-disable-notifications"):
            return False

        # Check notification age
        notified_age = self.get_days_since_date(self.tracked_updates['notified'], '%Y.%m.%d')
        if notified_age < self.settings.get_int("tracker-days-between-notifications"):
            self.logger.write("Tracker: Notification age is too small: %d days" % notified_age)
            return False

        notification_needed = False

        # Check maximum logged-in days
        if self.max_days >= self.settings.get_int("tracker-max-days"):
            self.logger.write("Tracker: Max days reached: %d days" % self.max_days)
            notification_needed = True
        else:
            max_age = self.get_days_since_date(self.oldest_since_date, '%Y.%m.%d')
            # Check maximum update age
            if max_age >= self.settings.get_int("tracker-max-age"):
                self.logger.write("Tracker: Max age reached: %d days" % max_age)
                notification_needed = True

        if not self.test_mode:
            # Check last time install button was pressed
            last_install_age = self.get_days_since_timestamp(self.settings.get_int("install-last-run"))
            if last_install_age <= self.settings.get_int("tracker-grace-period"):
                self.logger.write("Tracker: Mintupdate update button was pressed recently: %d days ago" % last_install_age)
                notification_needed = False
            else:
                # Check last time APT upgraded a package
                last_apt_upgrade = self.get_latest_apt_upgrade()
                last_apt_upgrade_age = self.get_days_since_date(last_apt_upgrade, '%Y-%m-%d')
                if last_apt_upgrade_age <= self.settings.get_int("tracker-grace-period"):
                    self.logger.write("Tracker: APT upgrades were taken recently: %d days ago" % last_apt_upgrade_age)
                    notification_needed = False

        if notification_needed:
            self.tracked_updates['notified'] = self.today
            return True
        else:
            return False

    # Records updates in JSON file and potentially notify
    def record(self):
        # Purge non-refreshed updates
        for name in list(self.tracked_updates['updates'].keys()):
            if name not in self.refreshed_update_names:
                del self.tracked_updates['updates'][name]
        # Update the check date
        self.tracked_updates['checked'] = self.today
        # Write JSON
        with open(self.path, "w") as f:
            json.dump(self.tracked_updates, f, indent=2)

try:
    gi.require_version('Flatpak', '1.0')
    from gi.repository import Flatpak
except:
    pass

class FlatpakUpdate():
    def __init__(self, op=None, installer=None, ref=None, installed_ref=None, remote_ref=None, pkginfo=None):
        if op is None:
            # json decoding
            return

        self.op = op

        # nullable
        self.installed_ref = installed_ref
        self.installing = self.installed_ref is None
        self.remote_ref = remote_ref
        self.pkginfo = pkginfo
        #

        # to be jsonified
        self.ref = ref
        self.ref_name = ref.get_name()
        self.metadata = op.get_metadata()
        self.size = op.get_download_size()
        self.link = installer.get_homepage_url(pkginfo) if pkginfo else None
        self.flatpak_type = "app" if ref.get_kind() == Flatpak.RefKind.APP else "runtime"
        self.old_version = ""
        self.new_version = ""
        self.name = ""
        self.summary = ""
        self.description = ""
        self.real_source_name = ""
        self.source_packages = []
        self.package_names = [self.ref_name]
        self.sub_updates = []
        self.origin = ""
        ##################

        # ideal:           old-version                     new-version
        # versions same:   old-version (commit)            new-version (commit)
        # no versions      commit                          commit

        # old version
        if installed_ref:
            iref_version = self.installed_ref.get_appdata_version()
            old_commit = installed_ref.get_commit()[:10]
        else:
            iref_version = ""
            old_commit = _("Installing")

        appstream_version = ""
        # new version
        if pkginfo:
            appstream_version = installer.get_version(pkginfo)

        new_commit = op.get_commit()[:10]

        if iref_version != "" and appstream_version != "":
            if iref_version != appstream_version:
                self.old_version = iref_version
                self.new_version = appstream_version
            else:
                self.old_version = "%s (%s)" % (iref_version, old_commit)
                self.new_version = "%s (%s)" % (appstream_version, new_commit)
        else:
            self.old_version = old_commit
            self.new_version = new_commit

        if pkginfo:
            self.name = pkginfo.get_display_name() or self.ref_name
        elif installed_ref and self.flatpak_type != "runtime":
            self.name = installed_ref.get_appdata_name()
        else:
            self.name = ref.get_name()

        # gnome has the branch as part of its name, the rest add it to the name if it's not a Locale
        if self.flatpak_type == "runtime":
            if ref.get_name() not in ("org.gnome.Platform", "org.gnome.Sdk") and (not ref.get_name().endswith(".Locale")):
                self.name = "%s (%s)" % (self.name, ref.get_branch())

        if pkginfo:
            self.summary = pkginfo.get_summary()
            if self.summary is None:
                if installed_ref:
                    self.summary = installed_ref.get_appdata_summary()

            self.description = installer.get_description(pkginfo)
            if self.description is None:
                self.description = self.summary
        else:
            self.summary = ""
            self.description = ""

        if self.summary == "" and self.flatpak_type == "runtime":
            self.summary = self.description = _("A Flatpak runtime package")

        self.real_source_name = self.ref_name
        self.source_packages = ["%s=%s" % (self.ref_name, self.new_version)]
        self.package_names = [self.ref_name]
        self.sub_updates = []

        if installed_ref:
            self.origin = installed_ref.get_origin().capitalize()
        elif remote_ref:
            self.origin = remote_ref.get_remote_name()
        else:
            self.origin = ""

    def add_package(self, update):
        self.sub_updates.append(update)
        self.package_names.append(update.name)
        self.size += update.size

    def to_json(self):
        trimmed_dict = {}

        for key in ("flatpak_type",
                    "name",
                    "origin",
                    "old_version",
                    "new_version",
                    "size",
                    "summary",
                    "description",
                    "real_source_name",
                    "source_packages",
                    "package_names",
                    "sub_updates",
                    "link",
                    "installing"):
            trimmed_dict[key] = self.__dict__[key]
        trimmed_dict["metadata"] = self.metadata.to_data()[0]
        trimmed_dict["ref"] = self.ref.format_ref()

        return trimmed_dict

    @classmethod
    def from_json(cls, json_data:dict):
        inst = cls()
        inst.flatpak_type = json_data["flatpak_type"]
        inst.ref = Flatpak.Ref.parse(json_data["ref"])
        inst.ref_name = inst.ref.get_name()
        inst.name = json_data["name"]
        inst.origin = json_data["origin"]
        inst.old_version = json_data["old_version"]
        inst.new_version = json_data["new_version"]
        inst.size = json_data["size"]
        inst.summary = json_data["summary"]
        inst.description = json_data["description"]
        inst.real_source_name = json_data["real_source_name"]
        inst.source_packages = json_data["source_packages"]
        inst.package_names = json_data["package_names"]
        inst.sub_updates = json_data["sub_updates"]
        inst.link = json_data["link"]
        inst.installing = json_data["installing"]
        inst.metadata = GLib.KeyFile()

        try:
            b = GLib.Bytes.new(json_data["metadata"].encode())
            inst.metadata.load_from_bytes(b, GLib.KeyFileFlags.NONE)
        except GLib.Error as e:
            print("flatpaks: unable to decode op metadata: %s" % e.message, file=sys.stderr, flush=True)

        return inst



#!/usr/bin/python3

import os
import subprocess
import time

if not os.path.exists("/var/lib/linuxmint/mintupdate-automatic-upgrades-enabled"):
    exit(0)

optionsfile = "/etc/mintupdate-automatic-upgrades.conf"
logfile = "/var/log/mintupdate.log"
power_connectfile="/sys/class/power_supply/AC/online"
log = open(logfile, "a")
log.write("\n-- Automatic Upgrade starting %s:\n" % time.strftime('%a %d %b %Y %H:%M:%S %Z'))
log.flush()

pkla_source = "/usr/share/linuxmint/mintupdate/automation/99-mintupdate-temporary.pkla"
pkla_target = "/etc/polkit-1/localauthority/90-mandatory.d/99-mintupdate-temporary.pkla"
try:
    power_supply_file = open(power_connectfile)
    powersupply = power_supply_file.read()[0]=='1'
    power_supply_file.close()
except:
    powersupply = True
    log.write(power_connectfile+" not found. Ignore power supply check.")
if powersupply:
    try:
        # Put shutdown and reboot blocker into place
        os.symlink(pkla_source, pkla_target)
    except:
        pass

    try:
        # Parse options file
        arguments = []
        if os.path.isfile(optionsfile):
            with open(optionsfile) as options:
                for line in options:
                    line = line.strip()
                    if line and not line.startswith("#"):
                        arguments.append(line)

        # Run mintupdate-cli through systemd-inhibit
        cmd = ["/bin/systemd-inhibit", '--why="Performing automatic updates"',
               '--who="Update Manager"',  "--what=shutdown", "--mode=block",
               "/usr/bin/mintupdate-cli", "upgrade", "--refresh-cache", "--yes"]
        cmd.extend(arguments)
        subprocess.run(cmd, stdout=log, stderr=log)

    except:
        import traceback
        log.write("Exception occurred:\n")
        log.write(traceback.format_exc())

    try:
        # Remove shutdown and reboot blocker
        os.unlink(pkla_target)
    except:
        pass

    log.write("-- Automatic Upgrade completed\n")
else:
    log.write("-- Power supply not connected, abort automatic update.\n")
log.close()



#!/usr/bin/python3

import apt
import aptkit.simpleclient
import configparser
import gettext
import gi
import os
import subprocess

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio
from Classes import _idle, _async
gettext.install("mintupdate", "/usr/share/locale")

class Assistant:

    def __init__(self):

        self.assistant = Gtk.Assistant()
        self.assistant.set_position(Gtk.WindowPosition.CENTER)
        self.assistant.set_title(_("System Upgrade"))
        self.assistant.connect("apply", self.apply_button_pressed)
        self.assistant.connect("cancel", self.cancel_button_pressed)
        self.assistant.connect("close", self.close_button_pressed)
        self.assistant.set_resizable(True)
        self.assistant.set_default_size(640, 480)

        # Intro page
        self.vbox_intro = Gtk.VBox()
        self.vbox_intro.set_border_width(60)
        page = self.assistant.append_page(self.vbox_intro)
        self.assistant.set_page_type(self.vbox_intro, Gtk.AssistantPageType.INTRO)
        self.assistant.set_page_title(self.vbox_intro, _("Introduction"))
        self.assistant.set_icon_name("mintupdate-release-upgrade")

        if not os.path.exists("/etc/linuxmint/info"):
            self.show_message('/usr/lib/linuxmint/mintUpdate/rel_upgrades/failure.png', _("Your system is missing critical components. A package corresponding to your edition of Linux Mint should provide the virtual package 'mint-info' and the file /etc/linuxmint/info."))
        else:
            self.current_codename = 'unknown'
            self.current_edition = 'unknown'
            self.edition = 'unknown'
            self.settings = None
            with open("/etc/linuxmint/info", "r") as info:
                for line in info:
                    line = line.strip()
                    if "EDITION=" in line:
                        self.current_edition = line.split('=')[1].replace('"', '').split()[0]
                        self.edition = self.current_edition.lower()
                    if "CODENAME=" in line:
                        self.current_codename = line.split('=')[1].replace('"', '').split()[0]

            rel_path = f"/usr/share/mint-upgrade-info/{self.current_codename}"
            if not os.path.exists(rel_path):
                self.show_message('/usr/lib/linuxmint/mintUpdate/rel_upgrades/info.png', _("No upgrades were found."))
            else:
                config = configparser.ConfigParser()
                config.read(f"{rel_path}/info")
                self.rel_target_name = config['general']['target_name']
                self.rel_target_codename = config['general']['target_codename']
                self.rel_editions = config['general']['editions']
                if self.edition in self.rel_editions:
                    label = Gtk.Label()
                    label.set_markup(_("A new version of Linux Mint is available!"))
                    self.vbox_intro.pack_start(label, False, False, 6)
                    image = Gtk.Image.new_from_file(f"{rel_path}/{self.edition}.png")
                    self.vbox_intro.pack_start(image, False, False, 0)
                    label = Gtk.Label()
                    label.set_markup("<b>%s</b>" % self.rel_target_name)
                    self.vbox_intro.pack_start(label, False, False, 0)
                    self.assistant.set_page_complete(self.vbox_intro, True)

                    # All good
                    self.meta = f"mint-meta-{self.edition}"
                    self.settings = None
                    self.screensaver_enabled = False
                    if self.edition == "cinnamon":
                        self.settings = Gio.Settings(schema="org.cinnamon.desktop.screensaver")
                        self.screensaver_enabled = self.settings.get_boolean("lock-enabled")
                    elif self.edition == "mate":
                        self.settings = Gio.Settings(schema="org.mate.screensaver")
                        self.screensaver_enabled = self.settings.get_boolean("lock-enabled")
                    self.build_assistant()
                else:
                    self.show_message('/usr/lib/linuxmint/mintUpdate/rel_upgrades/info.png', _("An upgrade was found but it is not available yet for the %s edition.") % self.current_edition)

        self.assistant.show_all()

    def build_assistant(self):
        # Known issues
        self.vbox_rel_notes = Gtk.VBox()
        self.vbox_rel_notes.set_border_width(60)
        self.assistant.append_page(self.vbox_rel_notes)
        self.assistant.set_page_title(self.vbox_rel_notes, _("Release notes"))
        self.assistant.set_page_type(self.vbox_rel_notes, Gtk.AssistantPageType.CONTENT)
        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file('/usr/lib/linuxmint/mintUpdate/rel_upgrades/info.png')
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(_("Please read the release notes before upgrading. They explain all the known issues, workarounds and solutions associated with the new version."))
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_rel_notes.pack_start(vbox_content, False, False, 6)
        link = Gtk.Label()
        link.set_markup("<a href='https://www.linuxmint.com/rel_%s.php'><b>%s</b></a>" % (self.rel_target_codename, _("Release notes for %s") % self.rel_target_name))
        self.vbox_rel_notes.pack_start(link, False, False, 6)
        label = Gtk.Label()
        label.set_markup("<i><b>%s</b></i>" % _("Click on the link to open the release notes."))
        self.vbox_rel_notes.pack_start(label, False, False, 6)
        self.assistant.set_page_complete(self.vbox_rel_notes, True)

        # New features
        self.vbox_new_features = Gtk.VBox()
        self.vbox_new_features.set_border_width(60)
        self.assistant.append_page(self.vbox_new_features)
        self.assistant.set_page_title(self.vbox_new_features, _("New features"))
        self.assistant.set_page_type(self.vbox_new_features, Gtk.AssistantPageType.CONTENT)
        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file('/usr/lib/linuxmint/mintUpdate/rel_upgrades/features.png')
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(_("Please look at the new features introduced in the new version."))
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_new_features.pack_start(vbox_content, False, False, 6)
        link = Gtk.Label()
        link.set_markup("<a href='https://www.linuxmint.com/rel_%s_whatsnew.php'><b>%s</b></a>" % (self.rel_target_codename, _("New features in %s") % self.rel_target_name))
        self.vbox_new_features.pack_start(link, False, False, 6)
        label = Gtk.Label()
        label.set_markup("<i><b>%s</b></i>" % _("Click on the link to browse the new features."))
        self.vbox_new_features.pack_start(label, False, False, 6)
        self.assistant.set_page_complete(self.vbox_new_features, True)

        # Warnings and risks
        self.vbox_prerequesites = Gtk.VBox()
        self.vbox_prerequesites.set_border_width(60)
        self.assistant.append_page(self.vbox_prerequesites)
        self.assistant.set_page_title(self.vbox_prerequesites, _("Requirements"))
        self.assistant.set_page_type(self.vbox_prerequesites, Gtk.AssistantPageType.CONFIRM)

        self.vbox_meta = Gtk.VBox()
        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file('/usr/lib/linuxmint/mintUpdate/rel_upgrades/failure.png')
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(_("The package %s needs to be installed before upgrading.") % self.meta)
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_meta.pack_start(vbox_content, False, False, 6)
        button = Gtk.Button()
        button.set_label(_("Install %s") % self.meta)
        packages = [self.meta]
        button.connect("button-release-event", self.install_pkgs, packages)
        self.vbox_meta.pack_start(button, False, False, 6)
        label = Gtk.Label()
        label.set_markup("<i><b>%s</b></i>" % _("Click on the button to install the missing package."))
        self.vbox_meta.pack_start(label, False, False, 6)
        self.vbox_meta.pack_start(Gtk.Separator(), False, False, 6)
        self.vbox_prerequesites.pack_start(self.vbox_meta, False, False, 10)

        if self.check_meta():
            self.vbox_meta.set_no_show_all(True)
            self.vbox_meta.hide()

        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file('/usr/lib/linuxmint/mintUpdate/rel_upgrades/risks.png')
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(_("New releases provide bug fixes and new features but they also sometimes introduce new issues. Upgrading always represents a risk. Your data is safe but new issues can potentially affect your operating system."))
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_prerequesites.pack_start(vbox_content, False, False, 6)
        self.check_button = Gtk.CheckButton()
        self.check_button.set_label(_("I understand the risk. I want to upgrade to %s.") % self.rel_target_name)
        self.check_button.connect("toggled", self.understood)
        self.vbox_prerequesites.pack_start(self.check_button, False, False, 6)

        self.assistant.set_page_complete(self.vbox_prerequesites, False)

        # Summary
        self.vbox_summary = Gtk.VBox()
        self.vbox_summary.set_border_width(60)
        self.assistant.append_page(self.vbox_summary)
        self.assistant.set_page_title(self.vbox_summary, _("Summary"))
        self.assistant.set_page_type(self.vbox_summary, Gtk.AssistantPageType.SUMMARY)


    def install_pkgs(self, widget, event, packages):
        client = aptkit.simpleclient.SimpleAPTClient(self.assistant)
        client.set_finished_callback(self.on_installation_finished)
        client.set_cancelled_callback(self.on_installation_finished)
        client.set_error_callback(self.on_installation_finished)
        client.install_packages(packages)

    def on_installation_finished(self, transaction=None, exit_state=None):
        self.check_reqs()

    def check_meta(self):
        cache = apt.Cache()
        if self.meta in cache:
            if cache[self.meta].is_installed:
                return True
        return False

    def understood(self, button):
        self.check_reqs()

    def check_reqs(self):
        if self.check_meta():
            self.vbox_meta.hide()
            if self.check_button.get_active():
                self.assistant.set_page_complete(self.vbox_prerequesites, True)
            else:
                self.assistant.set_page_complete(self.vbox_prerequesites, False)

    def show_message(self, icon, msg):
        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file(icon)
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(msg)
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_intro.pack_start(vbox_content, True, True, 0)
        self.assistant.set_page_complete(self.vbox_intro, False)

    def cancel_button_pressed(self, assistant):
        Gtk.main_quit()

    def close_button_pressed(self, assistant):
        Gtk.main_quit()

    def apply_button_pressed(self, assistant):
        # Turn off the screensaver during the upgrade
        if self.settings is not None:
            self.settings.set_boolean("lock-enabled", False)
        self.assistant.set_sensitive(False)
        self.launch_root_upgrade()

    @_async
    def launch_root_upgrade(self):
        subprocess.run(['pkexec', '/usr/bin/mint-release-upgrade-root', self.current_codename])
        self.show_result()

    @_idle
    def show_result(self):
        # Reset the screensaver the way it was before the upgrade
        if self.settings is not None:
            self.settings.set_boolean("lock-enabled", self.screensaver_enabled)

        self.assistant.set_sensitive(True)

        image_result = "failure"
        message_text = _("The upgrade did not succeed. Make sure you are connected to the Internet and try to upgrade again.")
        if os.path.exists("/etc/linuxmint/info"):
            with open("/etc/linuxmint/info", "r") as info:
                for line in info:
                    line = line.strip()
                    if "CODENAME=" in line:
                        new_codename = line.split('=')[1].replace('"', '').split()[0]
                        if new_codename == self.rel_target_codename:
                            image_result = "success"
                            message_text = _("Your operating system was successfully upgraded. Please reboot your computer for all changes to take effect.")
                            break

        vbox_content = Gtk.HBox()
        image = Gtk.Image.new_from_file(f'/usr/lib/linuxmint/mintUpdate/rel_upgrades/{image_result}.png')
        vbox_content.pack_start(image, False, False, 0)
        label = Gtk.Label()
        label.set_line_wrap(True)
        label.set_markup(message_text)
        vbox_content.pack_start(label, False, False, 6)
        self.vbox_summary.pack_start(vbox_content, False, False, 6)
        self.vbox_summary.show_all()
        self.assistant.set_page_complete(self.vbox_summary, True)

Assistant()
Gtk.main()



#!/usr/bin/python3

import sys, os
myPath = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, myPath + '/../usr/lib/linuxmint/mintUpdate/')

from datetime import datetime
from Classes import *
from mintUpdate import *
from checkAPT import *

# Test KernelVersion object
def test_kernel_version_series_comparison():
    version1 = KernelVersion ("4.8.0-2-generic")
    version2 = KernelVersion ("4.15.0-43-generic")
    version3 = KernelVersion ("4.15.0-44-generic")
    assert version1.version_id < version2.version_id
    assert version2.version_id < version3.version_id
    assert version1.series < version2.series
    assert version2.series == version3.series




#!/usr/bin/python3

import os
import json
import subprocess
import sys

import gi
gi.require_version('GLib', '2.0')
from gi.repository import GLib

try:
    if not GLib.find_program_in_path("flatpak"):
        raise Exception
    gi.require_version('Flatpak', '1.0')
    from gi.repository import Flatpak
except Exception:
    print("No Flatpak support - are flatpak and gir1.2-flatpak-1.0 installed?")
    raise NotImplementedError

from Classes import FlatpakUpdate

LOG_PATH = os.path.join(GLib.get_home_dir(), '.linuxmint', 'mintupdate', 'flatpak-updates.log')
UPDATE_WORKER_PATH = "/usr/lib/linuxmint/mintUpdate/flatpak-update-worker.py"

class FlatpakUpdater():
    def __init__(self):
        self.task = None

        self.updates = []
        self.error = None

    def refresh(self):
        self.kill_any_helpers()

        try:
            subprocess.run([UPDATE_WORKER_PATH, "--refresh"], timeout = 30, stdout=subprocess.DEVNULL)
        except subprocess.TimeoutExpired as e:
            print("Flatpaks: timed out trying to refresh", str(e))

    def fetch_updates(self):
        self.updates = []
        output = None

        try:
            output = subprocess.run([UPDATE_WORKER_PATH, "--fetch-updates"], timeout=45, stdout=subprocess.PIPE, encoding="utf-8").stdout
        except subprocess.TimeoutExpired as e:
            print("Flatpaks: timed out trying to get a list of updates", str(e))

        # print(output)

        if output is None:
            print("Flatpaks: no updates")
            return

        output = output.strip("\n")

        if output == "no-installed":
            print("Flatpaks: skipping update check - nothing installed")
            return

        if output.startswith("error:"):
            self.error = output[6:]
            print("Flatpaks: error from fetch-updates call", self.error)
            return

        try:
            json_data = json.loads(output)
            for item in json_data:
                self.updates.append(FlatpakUpdate.from_json(item))
        except json.JSONDecodeError as e:
            print("Flatpaks: unable to parse updates list", str(e))

        print("Flatpak: done generating updates")

    def prepare_start_updates(self, updates):
        argv = [UPDATE_WORKER_PATH, "--update-packages"] + [update.ref.format_ref() for update in updates]

        try:
            self.proc = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8")
        except Exception as e:
            print("Flatpak: Could not start worker: %s" % str(e))
            return False

        self.out_pipe = self.proc.stdout
        self.in_pipe = self.proc.stdin

        try:
            res = self.out_pipe.readline()
            if res:
                answer = res.strip("\n")
                if answer == "ready":
                    return
                else:
                    raise Exception("Unexpected response from worker - expected 'ready', got '%s'" % answer)
        except Exception as e:
            print("Flatpak: Could not set up to install updates: %s", str(e))
            self.terminate_helper()

    def perform_updates(self):
        try:
            self.in_pipe.write("start")
            self.in_pipe.flush()

            res = self.out_pipe.readline()
            if res:
                answer = res.strip("\n")
                if answer == "done":
                    print("Flatpaks: updates complete")
                elif answer.startswith("error:"):
                    self.error = answer[6:]
                    print("Flatpaks: error performing updates: %s" % self.error)
        except Exception as e:
            print("Flatpak: Could not perform updates: %s", str(e))

        self.terminate_helper()

    def terminate_helper(self):
        if self.proc is None:
            return

        try:
            self.in_pipe.close()
            self.out_pipe.close()
        except:
            pass

        self.proc.terminate()
        try:
            if self.error is not None:
                self.proc.wait() # Error dialog happening
            else:
                self.proc.wait(5)
        except subprocess.TimeoutExpired:
            self.proc.kill()

        self.proc = None

    def kill_any_helpers(self):
        try:
            os.system("killall -q flatpak-update-worker")
        except Exception as e:
            print (e)
            print(sys.exc_info()[0])












#!/usr/bin/python3
import os
import re
import sys

import apt

from Classes import (CONFIGURED_KERNEL_TYPE, SUPPORTED_KERNEL_TYPES,
                     KernelVersion, get_release_dates)

if len(sys.argv) > 1 and sys.argv[1] in SUPPORTED_KERNEL_TYPES:
    CONFIGURED_KERNEL_TYPE = sys.argv[1]

release_dates = None
try:
    current_version = os.uname().release
    cache = apt.Cache()
    signed_kernels = ['']
    local_kernels = {}
    r = re.compile(r'^(?:linux-image-)(?:unsigned-)?(\d.+?)(%s)$' % "|".join(SUPPORTED_KERNEL_TYPES))
    for pkg_name in cache.keys():
        installed = 0
        used = 0
        installable = 0
        pkg_version = ""
        pkg_match = r.match(pkg_name)
        if pkg_match:
            pkg = cache[pkg_name]
            pkg_data = None
            if pkg.candidate:
                pkg_data = pkg.candidate
            elif pkg.installed:
                pkg_data = pkg.installed
            else:
                continue
            version = pkg_match.group(1)
            kernel_type = pkg_match.group(2)
            full_version = version + kernel_type
            if pkg.is_installed:
                installed = 1
                pkg_version = pkg.installed.version
            else:
                # only offer to install same-type kernels
                if kernel_type != CONFIGURED_KERNEL_TYPE:
                    continue
                if pkg.candidate and pkg.candidate.downloadable:
                    installable = 1
                    pkg_version = pkg.candidate.version
            # filter duplicates (unsigned kernels where signed exists)
            if full_version in signed_kernels:
                continue
            signed_kernels.append(full_version)
            if full_version == current_version:
                used = 1

            # provide a representation of the version which helps sorting the kernels
            versions = KernelVersion(pkg_version).version_id

            if not pkg_data.origins[0].origin:
                origin = 0
            elif pkg_data.origins[0].origin == 'Ubuntu':
                origin = 1
            else:
                origin = 2

            archive = pkg_data.origins[0].archive

            # get support duration
            supported_tag = pkg_data.record.get("Supported")
            if not supported_tag and origin == 1 and "-proposed" not in pkg_data.origins[0].archive:
                # Workaround for Ubuntu releasing kernels by copying straight from
                # -proposed and only adding the Supported tag shortly after.
                # To avoid user confusion in the time in-between we just assume
                # that all Ubuntu kernels in all pockets but -proposed are supported
                # and generate the supported tag based on the distro support duration
                if not release_dates:
                    release_dates = get_release_dates()
                distro = pkg.candidate.origins[0].archive.split("-")[0]
                if distro in release_dates:
                    distro_lifetime = (release_dates[distro][1].year - release_dates[distro][0].year) * 12 +\
                                      release_dates[distro][1].month - release_dates[distro][0].month
                    if distro_lifetime >= 12:
                        supported_tag = "%sy" % (distro_lifetime // 12)
                    else:
                        supported_tag = "%sm" % distro_lifetime
            if supported_tag:
                if supported_tag.endswith("y"):
                    # override support duration for HWE kernels in LTS releases,
                    # these will be handled by the kernel window
                    if "-hwe" in pkg_data.source_name:
                        support_duration = -1
                    else:
                        support_duration = int(supported_tag[:-1]) * 12
                elif supported_tag.endswith("m"):
                    support_duration = int(supported_tag[:-1])
                else:
                    # unexpected support tag
                    support_duration = 0
            else:
                # unsupported
                support_duration = 0

            resultString = "KERNEL###%s###%s###%s###%s###%s###%s###%s###%s###%s###%s" % \
                (".".join(versions), version, pkg_version, installed, used, installable,
                    origin, archive, support_duration, kernel_type)
            print(resultString.encode("utf-8").decode('ascii', 'xmlcharrefreplace'))

except Exception as e:
    print("ERROR###ERROR###ERROR###ERROR")
    print("%s: %s\n" % (e.__class__.__name__, e), file=sys.stderr)
    sys.exit(1)



#!/usr/bin/python3

import os
import datetime
import tempfile
import traceback

class Logger():

    def __init__(self):
        self._create_log()
        self.callback = None

    def _create_log(self):
        path = os.path.join(tempfile.gettempdir(), "mintUpdate/")
        if not os.path.exists(path):
            os.umask(0)
            os.makedirs(path)
        self.log = tempfile.NamedTemporaryFile(mode="w", prefix=path, delete=False)
        try:
            os.chmod(self.log.name, 0o666)
        except:
            traceback.print_exc()

    def _log_ready(self):
        if self.log.closed:
            return False
        if not os.path.exists(self.log.name):
            self.log.close()
            self._create_log()
        return True

    def _write(self, separator, line):
        timestamp = datetime.datetime.now().strftime('%Y.%m.%d@%H:%M')
        line = f"{timestamp} {separator} {line}\n"
        if self._log_ready():
            self.log.write(line)
            self.log.flush()
        if self.callback:
            self.callback(line)

    def write(self, line):
        self._write("++", line)

    def write_error(self, line):
        self._write("--", line)

    def read(self):
        if not os.path.exists(self.log.name):
            self._create_log()
            return ""
        else:
            with open(self.log.name) as f:
                return f.read()

    def close(self):
        self.log.close()

    def set_callback(self, callback):
        self.callback = callback

    def remove_callback(self):
        self.callback = None



#!/usr/bin/python3

import argparse
import fnmatch
import os
import subprocess
import sys
import traceback

from checkAPT import APTCheck
from Classes import PRIORITY_UPDATES

if __name__ == "__main__":

    def is_blacklisted(blacklisted_packages, source_name, version):
        for blacklist in blacklisted_packages:
            if "=" in blacklist:
                (bl_pkg, bl_ver) = blacklist.split("=", 1)
            else:
                bl_pkg = blacklist
                bl_ver = None
            if fnmatch.fnmatch(source_name, bl_pkg) and (not bl_ver or bl_ver == version):
                return True
        return False

    parser = argparse.ArgumentParser(prog="mintupdate-cli")
    parser.add_argument("command", help="command to run (possible commands are: list, upgrade)")

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-k", "--only-kernel", action="store_true", help="only include kernel updates")
    group.add_argument("-s", "--only-security", action="store_true", help="only include security updates")
    parser.add_argument("-i", "--ignore", help="list of updates to ignore (comma-separated list of source package names). Note: You can also blacklist updates by adding them to /etc/mintupdate.blacklist, one source package per line. To ignore a specific version, use the format package=version.")
    parser.add_argument("-r", "--refresh-cache", action="store_true", help="refresh the APT cache")
    parser.add_argument("-d", "--dry-run", action="store_true", help="simulation mode, don't upgrade anything")
    parser.add_argument("-y", "--yes", action="store_true", help="automatically answer yes to all questions and always install new configuration files (unless you also use \"--keep-configuration\" option)")
    parser.add_argument("--install-recommends", action="store_true", help="install recommended packages")
    parser.add_argument("--keep-configuration", action="store_true", default=False, help="always keep local changes in configuration files (use with caution)")
    parser.add_argument("-v", "--version", action="version", version="__DEB_VERSION__", help="Display the current version")

    args = parser.parse_args()
    try:
        if args.refresh_cache:
            subprocess.run("sudo /usr/bin/mint-refresh-cache", shell=True)
        check = APTCheck()
        check.find_changes()

        blacklisted = []
        if os.path.exists("/etc/mintupdate.blacklist"):
            with open("/etc/mintupdate.blacklist") as blacklist_file:
                for line in blacklist_file:
                    line = line.strip()
                    if line.startswith("#"):
                        continue
                    blacklisted.append(line)
        if args.ignore:
            blacklisted.extend(args.ignore.split(","))

        updates = []
        for source_name in sorted(check.updates.keys()):
            update = check.updates[source_name]
            if source_name in PRIORITY_UPDATES:
                updates.append(update)
            elif args.only_kernel and update.type != "kernel":
                continue
            elif args.only_security and update.type != "security":
                continue
            elif is_blacklisted(blacklisted, update.real_source_name, update.new_version):
                continue
            else:
                updates.append(update)

        if args.command == "list":
            for update in updates:
                print ("%-15s %-45s %s" % (update.type, update.source_name, update.new_version))
        elif args.command == "upgrade":
            packages = []
            for update in updates:
                packages += update.package_names
            arguments = ["apt-get", "install"]
            if args.dry_run:
                arguments.append("--simulate")
            if args.yes:
                environment = os.environ
                environment.update({"DEBIAN_FRONTEND": "noninteractive"})
                arguments.append("--assume-yes")
                if not args.keep_configuration:
                    arguments.extend(["--option", "Dpkg::Options::=--force-confnew"])
            else:
                environment = None
            if args.install_recommends:
                arguments.append("--install-recommends")
            if args.keep_configuration:
                arguments.extend(["--option", "Dpkg::Options::=--force-confold"])
            subprocess.call(arguments + packages, env=environment)
    except:
        traceback.print_exc()
        sys.exit(1)



#!/usr/bin/python3

import apt
import aptkit.simpleclient
import gettext
import gi
import os
import subprocess
import sys

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

gettext.install("mintupdate", "/usr/share/locale")

if os.getuid() != 0:
    print("Run this code as root!")
    sys.exit(1)

if len(sys.argv) != 2:
    print("Missing arguments!")
    sys.exit(1)

codename = sys.argv[1]
sources_list = f"/usr/share/mint-upgrade-info/{codename}/official-package-repositories.list"
if not os.path.exists(sources_list):
    print("Unrecognized release: %s" % codename)
    sys.exit(1)

def file_to_list(filename):
    returned_list = []
    if os.path.exists(filename):
        with open(filename, 'r') as file_handle:
            for line in file_handle:
                line = line.strip()
                if line == "" or line.startswith("#"):
                    continue
                returned_list.append(line)
    return returned_list

class Upgrader():

    def __init__(self):
        self.refresh_done = False
        self.client = aptkit.simpleclient.SimpleAPTClient()
        self.additions = file_to_list(f"/usr/share/mint-upgrade-info/{codename}/additions")
        self.removals = file_to_list(f"/usr/share/mint-upgrade-info/{codename}/removals")
        self.blacklist = file_to_list(f"/usr/share/mint-upgrade-info/{codename}/blacklist")

    def run(self):
        self.update_cache()

    def update_cache(self):
        # STEP 1: UPDATE APT SOURCES
        if os.path.exists("/etc/apt/sources.list.d/official-source-repositories.list"):
            subprocess.run(["rm", "-f", "/etc/apt/sources.list.d/official-source-repositories.list"])
        subprocess.run(["cp", sources_list, "/etc/apt/sources.list.d/official-package-repositories.list"])

        # STEP 2: UPDATE APT CACHE
        self.client.update_cache()
        self.client.set_finished_callback(self.on_cache_updated)
        self.client.update_cache()

    def on_cache_updated(self, transaction=None, exit_state=None):
        # STEP 3: INSTALL MINT UPDATES
        cache = apt.Cache()
        cache.open(None)
        cache.upgrade(True)
        packages = []
        for pkg in cache.get_changes():
            # upgradeable..
            if pkg.is_installed and pkg.marked_upgrade:
                # not in blacklist..
                if pkg.candidate.source_name not in self.blacklist:
                    # with a new version..
                    if (pkg.candidate.version != pkg.installed.version):
                        # from the Mint repos..
                        for origin in pkg.candidate.origins:
                            if origin.origin == "linuxmint":
                                if origin.component != "romeo" and pkg.name != "linux-kernel-generic":
                                    packages.append(pkg.name)

        if len(packages) > 0:
            self.client.set_finished_callback(self.on_install_finished)
            self.client.install_packages(packages)
        else:
            self.on_install_finished()

    def on_install_finished(self, transaction=None, exit_state=None):
        # STEP 4: ADD PACKAGES
        if len(self.additions) > 0:
            self.client.set_finished_callback(self.on_additions_finished)
            self.client.install_packages(self.additions)
        else:
            self.on_additions_finished()

    def on_additions_finished(self, transaction=None, exit_state=None):
        # STEP 5: REMOVE PACKAGES
        if len(self.removals) > 0:
            self.client.set_finished_callback(self.on_removals_finished)
            self.client.remove_packages(self.removals)
        else:
            self.on_removals_finished()

    def on_removals_finished(self, transaction=None, exit_state=None):
        # STEP 6: UPDATE GRUB
        try:
            subprocess.run(["update-grub"])
            if os.path.exists("/usr/share/ubuntu-system-adjustments/systemd/adjust-grub-title"):
                subprocess.run(["/usr/share/ubuntu-system-adjustments/systemd/adjust-grub-title"])
        except Exception as detail:
            syslog.syslog("Couldn't update grub: %s" % detail)
        Gtk.main_quit()
        sys.exit(0)

if __name__ == "__main__":
    upgrader =Upgrader()
    upgrader.run()
    Gtk.main()


#!/usr/bin/python3

import sys
import apt_pkg

try:
    selection = sys.argv[1:]
    #print ' '.join(str(pkg) for pkg in selection)
    packages_to_install = []
    packages_to_remove = []

    apt_pkg.init()
    cache = apt_pkg.Cache(None)

    depcache = apt_pkg.DepCache(cache)
    depcache.init()

    with apt_pkg.ActionGroup(depcache):
        for package in selection:
            pkg = cache[package]
            #print "Marking : %s to install" % pkg.Name
            depcache.mark_install(pkg)

    depcache.fix_broken()

    #print "Install : %d" % depcache.inst_count
    #print "Remove : %d" % depcache.del_count

    # Get changes
    for pkg in cache.packages:
        if (not depcache.marked_keep(pkg) and
                (depcache.marked_install(pkg) or depcache.marked_upgrade(pkg)) and
                pkg.name not in selection and
                "%s:%s" % (pkg.name, pkg.architecture) not in selection and
                pkg.name not in packages_to_install):
            packages_to_install.append(pkg.name)
        if depcache.marked_delete(pkg) and pkg.name not in packages_to_remove:
            packages_to_remove.append(pkg.name)
    installations = ' '.join(packages_to_install)
    removals = ' '.join(packages_to_remove)
    print("%s###%s" % (installations, removals))
except Exception as e:
    print(e)
    print(sys.exc_info()[0])



#!/usr/bin/python3

import functools
import os
import argparse
import json
import sys
import setproctitle
from pathlib import Path
import signal
import locale
import gettext

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Flatpak', '1.0')
from gi.repository import Gtk, GLib, Flatpak, Gio

from mintcommon.installer import installer
from mintcommon.installer import _flatpak

from Classes import FlatpakUpdate

# i18n
APP = 'mintupdate'
LOCALE_DIR = "/usr/share/locale"
locale.bindtextdomain(APP, LOCALE_DIR)
gettext.bindtextdomain(APP, LOCALE_DIR)
gettext.textdomain(APP)
_ = gettext.gettext

setproctitle.setproctitle("flatpak-update-worker")
GLib.set_application_name(_("Flatpak updates"))

CHUNK_SIZE = 4096

LOG_PATH = os.path.join(GLib.get_home_dir(), '.linuxmint', 'mintupdate', 'flatpak-updates.log')

DEBUG_MODE = False
try:
    DEBUG_MODE = os.environ["DEBUG"]
except:
    pass

def debug(*args):
    if not DEBUG_MODE:
        return
    sanitized = [str(arg) for arg in args if arg is not None]
    argstr = " ".join(sanitized)
    print("flatpak-update-worker (DEBUG): %s" % argstr, file=sys.stderr, flush=True)

def warn(*args):
    sanitized = [str(arg) for arg in args if arg is not None]
    argstr = " ".join(sanitized)
    print("flatpak-update-worker (WARN): %s" % argstr, file=sys.stderr, flush=True)

class FlatpakUpdateWorker():
    def __init__(self):
        self.installer = installer.Installer(installer.PKG_TYPE_FLATPAK)
        self.fp_sys = _flatpak.get_fp_sys()
        self.task = None

        self.cancellable = Gio.Cancellable()

        if not self.check_for_any_installed():
            self.send_to_updater("no-installed")
            self.cancellable.cancel()
            self.quit()

        self.stdin = Gio.UnixInputStream.new(sys.stdin.fileno(), True)

        self.updates = []

    def check_for_any_installed(self):
        try:
            installed = self.fp_sys.list_installed_refs(self.cancellable)
        except GLib.Error as e:
            warn("not able to list installed refs: (%d) %s" % (e.code, e.message))
            installed = []

        if len(installed) == 0:
            debug("no flatpaks installed, exiting without refreshing")
            return False

        debug("%d flatpaks installed, continuing" % len(installed))
        return True

    def refresh(self, init=True):
        if self.cancellable.is_cancelled():
            return

        self.fp_sys.cleanup_local_refs_sync(None)
        self.fp_sys.prune_local_repo(None)

        if init:
            self.installer.init_sync()

        self.installer.force_new_cache_sync()

    def fetch_updates(self):
        if self.cancellable.is_cancelled():
            return

        self.installer.connect("appstream-changed", self.on_appstream_loaded)

        if not self.installer.init_sync():
            warn("cache not valid, refreshing")
            self.installer.init()
        else:
            debug("cache valid")

    def on_appstream_loaded(self, installer):
        self.updates = []
        self.installer.select_flatpak_updates(None,
                                              self._fetch_task_ready, self._fetch_updates_error, 
                                              None, None,
                                              use_mainloop=False)

    def _fetch_task_ready(self, task):
        debug("task object:", task, "transaction:", task.transaction)

        self.task = task
        self.error = task.error_message

        if self.error is None and task.transaction is not None:
            self._process_fetch_task(task)

            out = json.dumps(self.updates, default=lambda o: o.to_json(), indent=4)
            self.send_to_updater(out)
        else:
            self.send_to_updater(self.error)

        self.quit()
        debug("done generating updates", self.error)

    def _fetch_updates_error(self, task):
        warn("fetch error", task.error_message)

    def _process_fetch_task(self, task):
        trans = task.transaction
        ops = trans.get_operations()

        def cmp_ref_name(a, b):
            ref_a = Flatpak.Ref.parse(a.get_ref())
            ref_b = Flatpak.Ref.parse(b.get_ref())
            a_refstr_len = len(ref_a.get_name().split("."))
            b_refstr_len = len(ref_b.get_name().split("."))

            return a_refstr_len < b_refstr_len

        ops.sort(key=functools.cmp_to_key(cmp_ref_name))
        ops.sort(key=lambda op: Flatpak.Ref.parse(op.get_ref()).get_name())

        for op in ops:
            if op.get_operation_type() == Flatpak.TransactionOperationType.UPDATE:
                ref = Flatpak.Ref.parse(op.get_ref())
                debug("Update: ", op.get_ref(), ref.get_branch())
                try:
                    installed_ref = self.fp_sys.get_installed_ref(ref.get_kind(),
                                                                  ref.get_name(),
                                                                  ref.get_arch(),
                                                                  ref.get_branch(),
                                                                  None)
                    installed_ref.load_appdata()
                except GLib.Error as e:
                    if e.code == Flatpak.Error.NOT_INSTALLED:
                        installed_ref = None

                pkginfo = self.installer.find_pkginfo(ref.format_ref(), installer.PKG_TYPE_FLATPAK, remote=op.get_remote())
                try:
                    update = FlatpakUpdate(op, self.installer, ref, installed_ref, None, pkginfo)

                    if self.is_base_package(update) or (not self.add_to_parent_update(update)):
                        self.updates.append(update)
                except Exception as e:
                    warn("Problem creating FlatpakUpdate for %s: %s" % (ref.format_ref(), e))

            elif op.get_operation_type() == Flatpak.TransactionOperationType.INSTALL:
                ref = Flatpak.Ref.parse(op.get_ref())
                debug("Install: ", op.get_ref())
                try:
                    remote_ref = self.fp_sys.fetch_remote_ref_sync(op.get_remote(),
                                                                   ref.get_kind(),
                                                                   ref.get_name(),
                                                                   ref.get_arch(),
                                                                   ref.get_branch(),
                                                                   None)
                except GLib.Error as e:
                    debug("Can't add ref to install: %s" % e.message)
                    remote_ref = None

                pkginfo = self.installer.find_pkginfo(ref.format_ref(), installer.PKG_TYPE_FLATPAK, remote=op.get_remote())
                try:
                    update = FlatpakUpdate(op, self.installer, ref, None, remote_ref, pkginfo)

                    if self.is_base_package(update) or (not self.add_to_parent_update(update)):
                        self.updates.append(update)
                except Exception as e:
                    warn("Problem creating FlatpakUpdate for %s: %s" % (ref.format_ref(), e))
        task.cancel()

    def add_to_parent_update(self, update):
        for maybe_parent in self.updates:
            if not update.ref_name.startswith(maybe_parent.ref_name):
                continue

            kf = update.metadata

            try:
                extension_of = kf.get_string("ExtensionOf", "ref")
                if extension_of == maybe_parent.ref.format_ref():
                    maybe_parent.add_package(update)
                    return True
            except:
                pass

            try:
                runtime_ref_id = "runtime/%s" % kf.get_string("Runtime", "runtime")

                parent_ref = maybe_parent.ref.format_ref()
                if parent_ref == runtime_ref_id:
                    maybe_parent.add_package(update)
                    return True
            except:
                pass

    def is_base_package(self, update):
        name = update.ref.format_ref()
        if name.startswith("app"):
            return True
        try:
            kf = update.metadata
            runtime_ref_id = "runtime/%s" % kf.get_string("Runtime", "runtime")
            if name == runtime_ref_id:
                return True
        except Exception:
            return False

    def prepare_start_updates(self, updates):
        if self.cancellable.is_cancelled():
            return

        debug("creating real update task")
        self.error = None

        self.installer.select_flatpak_updates(updates,
                                              self._start_task_ready, self._start_updates_error,
                                              self._execute_finished, None)

    def _start_task_ready(self, task):
        self.task = task
        self.send_to_updater("ready")

        self.stdin.read_bytes_async(4096, GLib.PRIORITY_DEFAULT, None, self.message_from_updater)

    def _start_updates_error(self, task):
        warn("start updates error", task.error_message)
        self.send_to_updater(f"error:{task.error_message}")

    def start_updates(self):
        self.installer.execute_task(self.task)

    def _execute_finished(self, task):
        self.error = task.error_message
        if task.error_message is not None:
            self.send_to_updater(f"error:{task.error_message}")
        else:
            self.send_to_updater("done")
        self.write_to_log(task)
        self.quit()

    def write_to_log(self, task):
        try:
            entries = task.get_transaction_log()
        except:
            return

        directory = Path(LOG_PATH).parent

        try:
            os.makedirs(directory, exist_ok=True)
            with open(LOG_PATH, "a") as f:
                for entry in entries:
                    f.write("%s\n" % entry)
        except Exception as e:
            warn("Can't write to flatpak update log:", e)

    def send_to_updater(self, msg):
        print(msg, flush=True)

    def message_from_updater(self, pipe, res):
        if self.cancellable is None or self.cancellable.is_cancelled():
            return

        try:
            bytes_read = pipe.read_bytes_finish(res)
        except GLib.Error as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                warn("Error reading from updater: %s" % e.message)
            return

        if bytes_read:
            message = bytes_read.get_data().decode().strip("\n")
            debug("receiving from updater: '%s'" % message)

            if message == "confirm":
                self.confirm_start()
            elif message == "start":
                self.start_updates()

        pipe.read_bytes_async(4096, GLib.PRIORITY_DEFAULT, self.cancellable, self.message_from_updater)

    def quit(self, data=None):
        GLib.timeout_add(0, self.quit_on_ml)

    def quit_on_ml(self):
        if self.task:
            self.task.cancel()

        self.cancellable.cancel()
        Gtk.main_quit()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Flatpak worker for mintupdate")
    parser.add_argument("-d", "--debug", help="Print debugging information.",
                        action="store_true")
    parser.add_argument("-r", "--refresh", help="Refresh local flatpak cache and appstream info.",
                        action="store_true")
    parser.add_argument("-f", "--fetch-updates", help="Get a json list of update info.",
                        action="store_true")
    parser.add_argument("-u", "--update-packages", help="Updates packages - one or more flatpak ref strings must be supplied. "
                                                        "This process will remain running for communication.",
                        action="store_true")
    
    parser.add_argument('refs', metavar='ref', type=str, nargs='*', help='refs to update')
    args = parser.parse_args()

    updater = FlatpakUpdateWorker()
    GLib.unix_signal_add(GLib.PRIORITY_HIGH, signal.SIGTERM, updater.quit, None)

    if args.refresh:
        try:
            updater.refresh()
            exit(0)
        except Exception as e:
            print(e)
            exit(1)
    elif args.fetch_updates:
        updater.fetch_updates()
    elif args.update_packages:
        if len(args.refs) == 0:
            print("Expected one or more space-separated flatpak refs")
            exit(1)
        updater.prepare_start_updates(args.refs)
    else:
        print("nothing to do")
        exit(0)

    try:
        Gtk.main()
    except KeyboardInterrupt:
        Gtk.main_quit()

    exit(0)
